@@ -1,6 +1,7 @@
 from meteor_reasoner.automata.utils import *
-from meteor_reasoner.classes.atom import Atom
-import copy
+from meteor_reasoner.materialization.apply import *
+from meteor_reasoner.materialization.coalesce import *
+from collections import defaultdict
 
 
 def check_satisfy_dataset(w, D, involved_predicates=[]):
@@ -18,16 +19,10 @@ def check_satisfy_dataset(w, D, involved_predicates=[]):
     """
     for ruler_interval in w.ruler_intervals:
         for predicate in involved_predicates:
-            if type(D[predicate]) == list:
-                interval_list = D[predicate]
+            for entity, interval_list in D[predicate].items():
                 if interval_intesection_intervallist(ruler_interval, interval_list) and Atom(predicate) not \
                         in w.get_ruler_intervals_literals()[ruler_interval]:
                     return False
-            else:
-                for entity, interval_list in D[predicate].items():
-                    if interval_intesection_intervallist(ruler_interval, interval_list) and Atom(predicate) not \
-                            in w.get_ruler_intervals_literals()[ruler_interval]:
-                        return False
     return True
 
 
@@ -43,6 +38,7 @@ def check_satisfy_program(w, program):
         boolean.
 
     """
+
     ruler_intervals_literals = w.ruler_intervals_literals
     for ruler_interval in w.ruler_intervals:
         for rule in program:
@@ -201,6 +197,65 @@ def ground_rule(rule, installed_literals):
     return flag[0]
 
 
+def return_must_heads(rule, installed_literals):
+    """
+    Grounding the rule and then check whether the grounded head exists when all grounded literals in the body exists.
+    Args:
+        rule (a Rule instance):
+        installed_literals (a list of Literal or Atom instances):
+    Returns:
+        boolean
+    """
+    head_predicate = rule.head.get_predicate()
+    head_entity = rule.head.get_entity()
+    body_literals = rule.body
+
+    def ground_body(global_literal_index, context, heads):
+        if global_literal_index == len(body_literals):
+            if head_predicate == "Bottom":
+                heads.add(Atom("Bottom"))
+
+            ground_head_entity = []
+            for i, term in enumerate(head_entity):
+                if term.type == "constant":
+                    ground_head_entity.append(term)
+                else:
+                    ground_head_entity.append(context[term.name])
+
+            ground_head_entity = tuple(ground_head_entity)
+            ground_head = copy.deepcopy(rule.head)
+            ground_head.set_entity(ground_head_entity)
+            heads.add(ground_head)
+        else:
+            for _, tmp_context in ground_literal(body_literals[global_literal_index], context, installed_literals):
+                ground_body(global_literal_index+1, {**context, **tmp_context}, heads)
+
+    heads = set()
+    ground_body(0, {}, heads)
+    return heads
+
+
+def add_p_literals(p_literals, w):
+    tmp_D = defaultdict(lambda: defaultdict(list))
+    for tmp_ruler_interval in w.ruler_intervals:
+        for tmp_atom in w.ruler_intervals_literals[tmp_ruler_interval]:
+            if isinstance(tmp_atom, Atom):
+                tmp_D[tmp_atom.get_predicate()][tmp_atom.get_entity()].append(tmp_ruler_interval)
+    coalescing_d(tmp_D)
+    tmp_must_installed_literals = defaultdict(list)
+    for literal in p_literals:
+        tmp_T = apply(literal, tmp_D)
+        if len(tmp_T) != 0:
+            tmp_must_installed_literals[literal] = tmp_T
+
+    for ruler_interval in w.ruler_intervals:
+        for tmp_literal, tmp_intervals in tmp_must_installed_literals.items():
+            for tmp_interval in tmp_intervals:
+                if Interval.inclusion(ruler_interval, tmp_interval):
+                    w.ruler_intervals_literals[ruler_interval] |= set([tmp_literal])
+
+
+
 
 
 
