@@ -59,7 +59,14 @@ def parse_rule(line):
                 raise Exception("{} has an incorrect syntax!".format(literal_str))
             negative_body.append(literal)
 
-    rule = Rule(head_atom, literals, negative_body=negative_body)
+    ordered_literals = []
+    for literal in literals:
+        if isinstance(literal, BinaryLiteral):
+            ordered_literals.append(literal)
+            literals.remove(literal)
+    literals = sorted(literals, key=lambda item: len(item.get_entity()), reverse=True)
+    ordered_literals = ordered_literals + literals
+    rule = Rule(head_atom, ordered_literals, negative_body=negative_body)
     return rule
 
 
