@@ -1,13 +1,18 @@
 from meteor_reasoner.classes.literal import *
 from functools import reduce
-from collections import defaultdict
-from meteor_reasoner.classes.term import *
 from meteor_reasoner.classes.atom import *
 from meteor_reasoner.classes.interval import *
 import math
 from decimal import Decimal
 
 
+def build(w, max_x, min_x, left, right):
+    print("The w is: {}", w)
+    print("The maximum number:", max_x)
+    print("The mininum number:", min_x)
+    return left[2:], right[:-2]
+
+
 def construct_left_right_pattern(points, gcd):
     """
     This function build two dictionaries used for finding the next interval when
@@ -26,20 +31,15 @@ def construct_left_right_pattern(points, gcd):
     right_end_point = points[0] + gcd
 
     pattern = set()
-
+    pattern.add(left_end_point)
+    pattern.add(right_end_point)
     # the pattern's range [x, x+gcd], all other points will fall into this range
     for item in points:
-        tmp = item
-        if left_end_point <= tmp <= right_end_point:
-            pattern.add(tmp)
-        while True:
-            if tmp > right_end_point:
-                tmp -= gcd
-            else:
-                tmp += gcd
-            if left_end_point <= tmp <= right_end_point:
-                pattern.add(tmp)
-                break
+        a, b = divmod(item-left_end_point, gcd)
+        if a == 0:
+            continue
+        else:
+            pattern.add(b)
 
     pattern = list(pattern)
     pattern.sort()
@@ -74,6 +74,7 @@ def construct_left_right_pattern(points, gcd):
     return left_interval_search_dict, right_interval_search_dict
 
 
+
 def interval_intesection_intervallist(target, interval_list):
     """
     Check whether an Interval instance has an intersection with one of the interval_list
@@ -148,9 +149,74 @@ def get_initial_ruler_intervals(points, left_border, right_border, gcd):
         left_border (float): the most left minimum value  of the Window
         right_border(float):  the most right maximum value  of the Window
         gcd (float):
-
     Returns:
     """
+
+    points = list(points)
+    points.sort()
+    #
+    # left_end_point = points[0]
+    # right_end_point = points[0] -gcd
+    #
+    # pattern = set()
+    # pattern.add(left_end_point)
+    # pattern.add(right_end_point)
+    # # the pattern's range [x, x+gcd], all other points will fall into this range
+    # for item in points:
+    #     a, b = divmod(item-left_end_point, gcd)
+    #     if a == 0:
+    #         continue
+    #     else:
+    #         pattern.add(b)
+    #
+    # pattern = list(pattern)
+    # pattern.sort()
+    #
+    #
+    # left_initial_intervals = []
+    # for i in range(len(pattern)):
+    #     if i == len(pattern) - 1:
+    #         left_initial_intervals.append(
+    #             Interval(pattern[i], pattern[i], False, False))
+    #     else:
+    #         left_initial_intervals.append(
+    #             Interval(pattern[i], pattern[i], False, False))
+    #         left_initial_intervals.append(
+    #             Interval(pattern[i], pattern[i + 1], True, True))
+    #
+    #
+    # left_end_point = points[-1]-gcd
+    # right_end_point = points[-1]
+    #
+    # pattern = set()
+    # pattern.add(left_end_point)
+    # pattern.add(right_end_point)
+    # # the pattern's range [x, x+gcd], all other points will fall into this range
+    # for item in points:
+    #     a, b = divmod(right_end_point - item, gcd)
+    #     if a == 0:
+    #         continue
+    #     else:
+    #         pattern.add(left_end_point + b)
+    #
+    # pattern = list(pattern)
+    # pattern.sort()
+    #
+    # right_initial_intervals = []
+    # for i in range(len(pattern)):
+    #     if i == len(pattern) - 1:
+    #         right_initial_intervals.append(
+    #             Interval(pattern[i], pattern[i], False, False))
+    #     else:
+    #         right_initial_intervals.append(
+    #             Interval(pattern[i], pattern[i], False, False))
+    #         right_initial_intervals.append(
+    #             Interval(pattern[i], pattern[i + 1], True, True))
+    #
+    # return left_initial_intervals, right_initial_intervals
+    #
+    #
+    #
     initial_ruler_intervals = set()
     for point in points:
         left = point
@@ -164,7 +230,6 @@ def get_initial_ruler_intervals(points, left_border, right_border, gcd):
             if left_border <= right <= right_border:
                 initial_ruler_intervals.add(right)
             right += gcd
-
     initial_ruler_intervals = list(initial_ruler_intervals)
     initial_ruler_intervals.sort()
 
@@ -183,7 +248,17 @@ def get_initial_ruler_intervals(points, left_border, right_border, gcd):
             windows_interval_left_right_values.append(
                 Interval(initial_ruler_intervals[i], initial_ruler_intervals[i+1], True, True))
 
-    return windows_interval_len, windows_interval_left_right_values
+    left_interval = Interval(points[0], points[0], False, False)
+    left_index = windows_interval_left_right_values.index(left_interval)
+
+    right_interval = Interval(points[-1], points[-1], False, False)
+    right_index = windows_interval_left_right_values.index(right_interval)
+
+    return windows_interval_left_right_values[:left_index+1], windows_interval_left_right_values[right_index:]
+
+
+    #return windows_interval_len, windows_interval_left_right_values
+
 
 
 def get_dataset_points_x(D, min_x_flag=False):
